{{ template "chrome/header.html" . }}
<body>
{{ template "chrome/navbar.html" . }}
<div class="container">
    <div class="row">
        <div class="col-md-9">
            <div class="well well-sm">
                <strong>Items with tag <code>{{ .Title | lower }}</code></strong>
                <ul class="list-unstyled">
                    {{ range .Data.Pages }}
                        {{ .Render "li" }}
                    {{ end}}
                </ul>
            </div>
        </div>

        <!-- Sidebar -->
        <div class="col-md-3">
        {{ template "chrome/menu.html" . }}
        </div>
    </div>
{{ template "chrome/footer.copyright.html" . }}
</div>
{{ template "chrome/footer.html" . }}