<!doctype html>
<html lang="en">
<head>
    {{ template "chrome/meta.html" . }}
    <title>{{ .Title }}</title>
    <link rel="canonical" href="{{ .Permalink }}">
    {{ template "chrome/header.includes.html" . }}
    {{ if .RSSlink }}<link href="{{ .RSSlink }}" rel="alternate" type="application/rss+xml" title="{{ .Site.Title }}" />{{ end }}
</head>
